import { extractPuppeteerRequest } from './request';
import { extractPuppeteerResponse } from './response';
import logger from '../logger';
import { loggerTrace, getTimeStampWithStrictOrder } from '../utils';
/**
 * There is no valid reason to have more than one driver instances per page
 */
const pagesSet = new Set();
class PuppeteerDriver {
    constructor({ page }) {
        logger.debug(`instantiating new driver`);
        if (!page) {
            throw new Error('"page" is not defined');
        }
        if (pagesSet.has(page)) {
            logger.error(`second driver instantiation on the same page`);
            logger.error(`(probably you did not stop the mocker before start it again)`);
            throw new Error(`second driver instantiation on the same page`);
        }
        this.page = page;
        this.setRequestInterception(true);
        pagesSet.add(page);
        this.page.on('close', () => {
            this.setRequestInterception(false);
            pagesSet.delete(this.page);
        });
    }
    async setRequestInterception(arg) {
        await this.page.setRequestInterception(arg);
    }
    async onRequest(fn) {
        const handler = async (interceptedRequest) => {
            const timestampWithOrder = getTimeStampWithStrictOrder();
            loggerTrace(`${interceptedRequest.url()} ← page.on('request') fired`);
            const { request, abort, next, respond } = await extractPuppeteerRequest(interceptedRequest, timestampWithOrder);
            fn({ request, abort, next, respond });
        };
        // Intercepting all requests and respinding with mocks
        this.page.on('request', handler);
        return async () => {
            await this.setRequestInterception(false);
            pagesSet.delete(this.page);
            this.page.off('request', handler);
        };
    }
    async onResponse(fn) {
        const handler = async (interceptedResponse) => {
            const timestampWithOrder = getTimeStampWithStrictOrder();
            const url = interceptedResponse.request().url();
            loggerTrace(`${url} → page.on('response') fired`);
            await fn(await extractPuppeteerResponse(interceptedResponse, timestampWithOrder));
            loggerTrace(`${url} → finish`);
        };
        // Intercepting all requests and respinding with mocks
        this.page.on('response', handler);
        return () => this.page.off('response', handler);
    }
    onClose(fn) {
        this.page.on('close', fn);
        return () => this.page.off('close', fn);
    }
}
export default PuppeteerDriver;
