// https://playwright.dev/docs/network
import { extractPlaywrightRequest } from './request';
import { extractPlaywrightResponse } from './response';
import logger from '../logger';
import { loggerTrace, getTimeStampWithStrictOrder } from '../utils';
import { URL } from 'url';
/**
 * There is no valid reason to have more than one driver instances per page
 */
const pagesSet = new Set();
const redirectCodes = [301, 302, 303, 305, 307, 308];
class PlaywrightDriver {
    constructor({ page, noParseResponseUrls = [] }) {
        this._routeUrl = () => true;
        logger.debug(`instantiating new playwright driver`);
        if (!page) {
            throw new Error('"page" is not defined');
        }
        if (pagesSet.has(page)) {
            logger.error(`second driver instantiation on the same page`);
            logger.error(`(probably you did not stop the mocker before start it again)`);
            throw new Error(`second driver instantiation on the same page`);
        }
        this.seenRedirects = new Map();
        this.noParseResponseUrls = noParseResponseUrls;
        this.page = page;
        pagesSet.add(page);
        this.page.on('close', () => {
            pagesSet.delete(this.page);
        });
    }
    async setRequestInterception(_arg) {
        // do nothing for playwright
    }
    async onRequest(fn) {
        const handler = async (route) => {
            const timestampWithOrder = getTimeStampWithStrictOrder();
            loggerTrace(`${route.request().url()} ← page.on('request') fired`);
            const { request, abort, next, respond } = extractPlaywrightRequest(route, timestampWithOrder);
            await fn({ request, abort, next, respond });
        };
        await this.page.route(this._routeUrl, handler);
        return async () => {
            pagesSet.delete(this.page);
            try {
                await this.page.unroute(this._routeUrl, handler);
            }
            catch (e) {
                if (e.message.indexOf('has been closed') === -1) {
                    throw e;
                }
            }
        };
    }
    async onResponse(fn) {
        const handler = async (interceptedResponse) => {
            const timestampWithOrder = getTimeStampWithStrictOrder();
            const url = interceptedResponse.request().url();
            if (redirectCodes.indexOf(interceptedResponse.status()) !== -1) {
                // see https://playwright.dev/docs/api/class-page#pagerouteurl-handler
                // Because of the way the page.route method works, the following redirection will not be handled by it.
                const newLocation = interceptedResponse.headers().location;
                const destUrl = new URL(newLocation, url).toString();
                this.seenRedirects.set(destUrl, (this.seenRedirects.get(destUrl) || 0) + 1);
                logger.debug(`Redirection to "${newLocation}" from "${url}" will not be handled.`);
            }
            if (this.seenRedirects.get(url)) {
                this.seenRedirects.set(url, this.seenRedirects.get(url) - 1);
                return;
            }
            const noParseResponse = this.noParseResponseUrls.some((urlPart) => url.indexOf(urlPart) !== -1);
            loggerTrace(`${url} → page.on('response') fired`);
            await fn(await extractPlaywrightResponse(interceptedResponse, {
                ...timestampWithOrder,
                noParseResponse
            }));
            loggerTrace(`${url} → finish`);
        };
        // Intercepting all requests and respinding with mocks
        await this.page.on('response', handler);
        return () => this.page.off('response', handler);
    }
    onClose(fn) {
        this.page.on('close', fn);
        return () => this.page.off('close', fn);
    }
}
export default PlaywrightDriver;
